package util;

import util.SmartStringArray;

import java.lang.reflect.Array;
import java.io.IOException;
import java.util.Stack;

import org.xml.sax.XMLReader;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

import org.apache.xerces.parsers.SAXParser;

/**
 * A JavaBean which uses SAX2 for extracting data
 * from an XML file. The SAX2 callbacks are defined
 * within local classes inside of the bean's getter 
 * methods. The general structure is as follows:
 * <pre>
 * public class XMLUtilsBean
 * {
 *   ...private fields...
 *   ...public constructor...
 *   public String[] getXXX()
 *   {
 *     ...local vars...
 *     class Handler extends DefaultHandler
 *     {
 *       ...SAX 2.0 callbacks...
 *     }
 *     ...parsing operations...
 *     ...Handler, ErrHandler referenced here... 
 *   }
 *   class ErrHandler extends DefaultHandler
 *   {
 *     ...
 *   }
 * }
 * </pre>
 *
 * @author Armen B. Ezekielian
 * @version 1.0
 */
public class XMLUtilsBean
{
  // The URI of the xml file to be processed. 
  private String XMLFileName; 

  // The name of the element to match against during 
  // processing of the xml file. 
  private String elementName; 

  // The name of the attribute to match against during 
  // processing of the xml file.
  private String attName; 

  // The value of the attribute to match against during
  // processing of the xml file.
  private String attValue; 

  // The data (parsed character data) for all
  // elements matched by elementName
  private String[] data;

  // The data (parsed character data) for all
  // the elements matched by the attName-attValue pair
  private String[] dataByAttValPair;

  // The attribute values for the given elementName and attName
  private String[] attValsByElement;

  // Lighter-weight alternative to Vector for storing element
  // content
  private SmartStringArray ssa;

  // boolean flags to determine if user-supplied XML names
  // and values are actually present in the XML file. If
  // not, the appropriate diagnostic message is returned.
  private boolean hasElementName;
  private boolean hasAttName;
  private boolean hasAttValue;

  /**
   * Requisite no-argument constructor. Performs initialization
   * of instance variables. The name of the XML file to be 
   * parsed should be properly set before referenced by methods.
   * This constructor sets the initial value of 
   * <code>XMLFileName</code> to <code>&quot;NOT_SET&quot;</code>.
   * All <code>getXXX</code> methods check if <code>XMLFileName</code>
   * is <code>&quot;NOT_SET&quot;</code>. If so, an 
   * <code>IOException</code> is thrown, if not, 
   * <code>XMLReader.parse(InputSource XMLFileName)</code>
   * is invoked.
   */
  public XMLUtilsBean()
  {
    super();

    XMLFileName = "NOT_SET"; // used by all getXXX methods
    elementName = ""; // used by all getXXX methods
    attName = ""; // used by some getXXX methods
    attValue = ""; // used by some getXXX methods

  } // XMLUtilsBean():constructor

  /**
   * Retrieves the name of the XML file being processed. 
   *
   * @return A <code>String</code> containing the name of the XML file
   *         being parsed. 
   * @see #setXMLFileName
   */
  public String getXMLFileName()
  {
    return XMLFileName;
  }

  /**
   * Sets the name of the XML file to be processed. 
   * This method is invoked directly or via a
   * <code>&lt;jsp:setProperty&gt;</code> bean tag
   * in a JSP page.
   *
   * @param name A <code>String</code> containing the name of the 
   *             XML file. 
   */
  public void setXMLFileName(String name)
  {
    XMLFileName = name;
  }

  /**
   * Retrieves the name of the desired XML tag.
   *
   * @return A <code>String</code> containing the name of the XML tag. 
   * @see #setElementName
   */
  public String getElementName()
  {
    return elementName;
  }

  /**
   * Sets the name of the desired XML tag.
   * This method is invoked directly or via
   * a <code>&lt;jsp:setProperty&gt;</code> bean tag
   * in a JSP page.
   *
   * @param name A <code>String</code> containing the name of the 
   *             XML element name
   */
  public void setElementName(String name)
  {
    elementName = name;
  }
  
  /**
   * Retrieves the name of the desired XML tag attribute name.
   *
   * @return A <code>String</code> containing the name of the XML tag 
   *         attribute. 
   * @see #setAttName
   */
  public String getAttName()
  {
    return attName;
  }

  /**
   * Sets the name of the desired XML tag attribute name.
   * This method is invoked directly or via a
   * <code>&lt;jsp:setProperty&gt;</code> bean tag
   * in a JSP page.
   *
   * @param name A <code>String</code> containing the name of the XML tag 
   *             attribute name.
   */
  public void setAttName(String name)
  {
    attName = name;
  }

  /**
   * Retrieves the name of the desired XML tag attribute value.
   *
   * @return A <code>String</code> containing the name of the XML 
   *         tag attribute value.
   * @see #setAttValue
   */
  public String getAttValue()
  {
    return attValue;
  }

  /**
   * Sets the name of the desired XML tag attribute value.
   * This method is invoked directly or via a
   * <code>&lt;jsp:setProperty&gt;</code> bean tag
   * in a JSP page.
   *
   * @param name A <code>String</code> containing the name of the XML tag 
   *             attribute value.
   */
  public void setAttValue(String name)
  {
    attValue = name;
  }

  /**
   * Returns the <code>(#PCDATA)</code> from all XML elements
   * with the given element tag name. Since this method references
   * the element name, that value is typically set prior to its 
   * usage via {@link #setElementName} or using a 
   * <code>&lt;jsp:setProperty&gt;</code> bean tag in a JSP page.
   *
   * @return A <code>String</code> array containing the data 
   * @throws org.xml.sax.SAXException
   *         Any SAX exception, possibly wrapping another 
   *         exception
   * @see #setElementName
   */
  public String[] getData() throws SAXException
  {
    boolean isEmpty = false;

    class Handler extends DefaultHandler
    {
      // Holds the content of characters() output
      private StringBuffer content;

      // For retrieving a snapshot of the current element context
      private Stack context;

      Handler()
      {
        super();
        content = new StringBuffer(); 
        ssa = new SmartStringArray(); 
        context = new Stack();
      }
   
      public void ignorableWhiteSpace(char ch[], int start, int length)
        throws SAXException
      {

      }
      
      public void startElement(String uri, String name,
        String qName, Attributes atts) throws SAXException
                             
      {
        ElementDetails details = new ElementDetails(qName, atts);
        context.push(details);
        content.setLength(0);
      }

      public void characters(char[] ch, int start, int length)
        throws SAXException
      {
        content.append(ch, start, length);
      }

      public void endElement(String uri, String name, 
        String qName) throws SAXException
      {
        char[] charData = content.toString().toCharArray();
        boolean isEmptyData = false;
        if(qName.equals(elementName))
        {
          hasElementName = true;
          int count = 0;
          for(int i = 0; i < Array.getLength(charData); i++)
          {
            if(Character.isWhitespace(charData[i]))
              count++;
          }
          if(Array.getLength(charData) == count) isEmptyData = true;
          if(!isEmptyData)
            ssa.add(content.toString());
        }
        content.setLength(0);
        context.pop();
      }
    } // Handler:class

    // Perform the parsing operations
    XMLReader xr = new SAXParser();
    xr.setContentHandler(new Handler());
    xr.setErrorHandler(new ErrHandler());
    try
    {
      xr.setFeature("http://xml.org/sax/features/validation", true);
      xr.setFeature("http://xml.org/sax/features/namespaces", false);
      xr.setFeature("http://xml.org/sax/features/namespace-prefixes", false);
      if(!XMLFileName.equals("NOT_SET"))
      {
        xr.parse(new InputSource(XMLFileName));
      }
      else
      {
        throw new IOException("XMLFileName not set");
      }
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe.getMessage());
    }
    catch(SAXNotSupportedException snse)
    {
      Exception ex = snse;
      if(snse.getException() != null)      
        ex = snse.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXNotRecognizedException snre)
    {
      Exception ex = snre;
      if(snre.getException() != null)      
        ex = snre.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXParseException spe)
    {
      Exception ex = spe;
      if(spe.getException() != null)      
        ex = spe.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXException se)
    {
      Exception ex = se;
      if(se.getException() != null)      
        ex = se.getException();
      throw new SAXException(ex.getMessage());
    }

    // Check if any element is non-empty. If all elements are
    // empty, including any amount of pure whitespace, return 
    // DATA_NOT_FOUND.
    data = ssa.toArray();
    int numData = Array.getLength(data);
    if(numData == 0) isEmpty = true;
    
    if(!hasElementName)
    {
      //return new String[] {"Element \"" + elementName +
      //                     "\" not present in " + XMLFileName}; 
      return new String[] {"TAG_NOT_FOUND"};
    }
    else if(isEmpty)
    { 
      hasElementName = false;
      return new String[] {"DATA_NOT_FOUND"};
    } 
    else 
    {
      hasElementName = false;
      return data;
    }
      
  } // getData():String[]

  /**
   * Returns all attribute values for a given element name
   * and having a given attribute name. Since this method references
   * both an element name and an attribute name, those values are 
   * typically set prior to its usage via {@link #setElementName} and 
   * {@link #setAttName} or using a 
   * <code>&lt;jsp:setProperty&gt;</code> bean
   * tag in a JSP page.
   *
   * @return A <code>String</code> array containing the data
   * @throws org.xml.sax.SAXException
   *         Any SAX exception, possibly wrapping another
   *         exception
   * @see #setElementName
   * @see #setAttName
   */
  public String[] getAttValsByElement() throws SAXException
  {
    class Handler extends DefaultHandler
    {
      // Holds the attribute values
      private StringBuffer content;

      // For retrieving a snapshot of the current element context
      private Stack context;

      Handler()
      {
        super();
        content = new StringBuffer();
        ssa = new SmartStringArray(); 
        context = new Stack();
      }
   
      public void ignorableWhiteSpace(char ch[], int start, int length)
        throws SAXException
      {

      }
      
      public void startElement(String uri, String name,
        String qName, Attributes atts) throws SAXException
                             
      {
        ElementDetails details = new ElementDetails(qName, atts);
        context.push(details);
	content.setLength(0);
      }

      public void characters(char[] ch, int start, int length)
        throws SAXException
      {

      }

      public void endElement(String uri, String name, 
        String qName) throws SAXException
      {
        if(qName.equals(elementName))
        {
          hasElementName = true;
          ElementDetails ed = (ElementDetails)context.peek();
          int attIndex = ed.atts.getIndex(attName);
          if(attIndex != -1)
          {
          hasAttName = true;
          content.append(ed.atts.getValue(attIndex));
          ssa.add(content.toString());
          }
        }
        content.setLength(0);
        context.pop();
      }
    } // Handler:class

    // Perform the parsing operations
    XMLReader xr = new SAXParser();
    xr.setContentHandler(new Handler());
    xr.setErrorHandler(new ErrHandler());
    try
    {
      xr.setFeature("http://xml.org/sax/features/validation", true);
      xr.setFeature("http://xml.org/sax/features/namespaces", false);
      xr.setFeature("http://xml.org/sax/features/namespace-prefixes", false);
      if(!XMLFileName.equals("NOT_SET"))
      {
        xr.parse(new InputSource(XMLFileName));
      }
      else
      {
        throw new IOException("XMLFileName not set");
      }
    }
    catch(SAXNotSupportedException snse)
    {
      Exception ex = snse;
      if(snse.getException() != null)      
        ex = snse.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXNotRecognizedException snre)
    {
      Exception ex = snre;
      if(snre.getException() != null)      
        ex = snre.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXParseException spe)
    {
      Exception ex = spe;
      if(spe.getException() != null)      
        ex = spe.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXException se)
    {
      Exception ex = se;
      if(se.getException() != null)      
        ex = se.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe.getMessage());
    }

    attValsByElement = ssa.toArray();
    
    if(!hasElementName)
    {
      return new String[] {"TAG_NOT_FOUND"};
    }
    else if(!hasAttName)
    {
      return new String[] {"ATTRIBUTE_NOT_FOUND"};
    }
    else 
    {
      hasElementName = false;
      hasAttName = false; 
      return attValsByElement;
    }
  } // getAttValsByElement():String[]

  /**
   * Returns the <code>(#PCDATA)</code> from all XML elements
   * having a given attribute name-value pair. Since this method 
   * references
   * both an attribute name and value, those values are
   * typically set prior to its usage via {@link #setAttName} and
   * {@link #setAttValue} or using a 
   * <code>&lt;jsp:setProperty&gt;</code> bean
   * tag in a JSP page.
   *
   * @return A <code>String</code> array containing the data
   * @throws org.xml.sax.SAXException
   *         Any SAX exception, possibly wrapping another
   *         exception
   * @see #setAttName
   * @see #setAttValue
   */
  public String[] getDataByAttValPair() throws SAXException
  {
    boolean isEmpty = false;

    class Handler extends DefaultHandler
    {
      // Holds the content of characters() output
      private StringBuffer content;

      // For retrieving a snapshot of the current element context
      private Stack context;

      Handler()
      {
        super();
        content = new StringBuffer(); 
        ssa = new SmartStringArray(); 
        context = new Stack();
      }
   
      public void ignorableWhiteSpace(char ch[], int start, int length)
        throws SAXException
      {

      }
      
      public void startElement(String uri, String name,
        String qName, Attributes atts) throws SAXException
                             
      {
        ElementDetails details = new ElementDetails(qName, atts);
        context.push(details);
        content.setLength(0);
      }

      public void characters(char[] ch, int start, int length)
        throws SAXException
      {
        content.append(ch, start, length);
      }

      public void endElement(String uri, String name, 
        String qName) throws SAXException
      {
        char[] charData = content.toString().toCharArray();
        String attVal = "";
        boolean isEmptyData = false;
        ElementDetails ed = (ElementDetails)context.peek();
        int attIndex = ed.atts.getIndex(attName);
        if(attIndex != -1)
        {
          hasAttName = true;
          attVal = ed.atts.getValue(attIndex);
          if(attVal.equals(attValue))
          {
            hasAttValue = true;
            int count = 0;
            for(int i = 0; i < Array.getLength(charData); i++)
            {
              if(Character.isWhitespace(charData[i]))
                count++;
            }
            if(Array.getLength(charData) == count) isEmptyData = true;
            if(!isEmptyData) ssa.add(content.toString());
          }
        }
        content.setLength(0);
        context.pop();
     }
    } // Handler:class

    // Perform the parsing operations
    XMLReader xr = new SAXParser();
    xr.setContentHandler(new Handler());
    xr.setErrorHandler(new ErrHandler());
    try
    {
      xr.setFeature("http://xml.org/sax/features/validation", true);
      xr.setFeature("http://xml.org/sax/features/namespaces", false);
      xr.setFeature("http://xml.org/sax/features/namespace-prefixes", false);
      if(!XMLFileName.equals("NOT_SET"))
      {
        xr.parse(new InputSource(XMLFileName));
      }
      else
      {
        throw new IOException("XMLFileName not set");
      }
    }
    catch(SAXNotSupportedException snse)
    {
      Exception ex = snse;
      if(snse.getException() != null)      
        ex = snse.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXNotRecognizedException snre)
    {
      Exception ex = snre;
      if(snre.getException() != null)      
        ex = snre.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXParseException spe)
    {
      Exception ex = spe;
      if(spe.getException() != null)      
        ex = spe.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(SAXException se)
    {
      Exception ex = se;
      if(se.getException() != null)      
        ex = se.getException();
      throw new SAXException(ex.getMessage());
    }
    catch(IOException ioe)
    {
      throw new SAXException(ioe.getMessage());
    }

    dataByAttValPair = ssa.toArray();
    int numData = Array.getLength(dataByAttValPair);
    if(numData == 0) isEmpty = true;
    
    if(!hasAttName)
    {
      return new String[] {"ATTRIBUTE_NOT_FOUND"};
    }
    else if(!hasAttValue)
    {
      return new String[] {"VALUE_NOT_FOUND"};
    }
    else if(isEmpty)
    { 
      hasAttName = false; 
      hasAttValue = false; 
      return new String[] { "DATA_NOT_FOUND" };
    } 
    else 
    {
      hasAttName = false; 
      hasAttValue = false; 
      return dataByAttValPair;
    }
  } // getDataByAttValPair():String[]

  class ErrHandler extends DefaultHandler
  {
    ErrHandler()
    { 
      super();
    }

    /* Warning */
    public void warning(SAXParseException spe) throws SAXException
    {
      String message = "**Parsing Warning** " + 
                       getLocationString(spe) + 
                       ": " + spe.getMessage();
      throw new SAXException(message);
    }

    /* Error */
    public void error(SAXParseException spe) throws SAXException
    {
      String message = "**Parsing Error** " + 
                       getLocationString(spe) + 
                       ": " + spe.getMessage();
      throw new SAXException(message);
    }

    /* Fatal error */
    public void fatalError(SAXParseException spe)
      throws SAXException
    {
      String message = "**Parsing Error** " + getLocationString(spe) + 
                ": " + spe.getMessage();
      throw new SAXException(message);
    }

    /** Returns a string of the location. */
    private String getLocationString(SAXParseException spe)
    {
      StringBuffer str = new StringBuffer();

      String systemId = spe.getSystemId();
      if (systemId != null)
      {
        int index = systemId.lastIndexOf('/');
        if (index != -1)
          systemId = systemId.substring(index + 1);
        str.append(systemId);
      }
      str.append(':');
      str.append(spe.getLineNumber());;
      str.append(':');
      str.append(spe.getColumnNumber());

      return str.toString();
    } // getLocationString(SAXParseException):String 
  } // class ErrHandler

  private class ElementDetails
  {
    public String qName;
    public Attributes atts;
    public ElementDetails(String qName, Attributes atts)
    {
      this.qName = qName;
      this.atts = new AttributesImpl(atts); // make a copy
    }
  }
} // public class XMLUtilsBean
