// Copyright (C) 1999 by Jason Hunter <jhunter@acm.org>.  All rights reserved.
// Use of this class is limited.  Please see the LICENSE for more information.

package com.oreilly.servlet;

import java.io.*;
import java.util.*;
import javax.servlet.*;

/** 
 * A class to simplify parameter handling.  It can return parameters of
 * any primitive type (no casting or parsing required), can throw an 
 * exception when a parameter is not found (simplifying error handling),
 * and can accept default values (eliminating error handling).
 * <p>
 * It is used like this:
 * <blockquote><pre>
 * ParameterParser parser = new ParameterParser(req);
 * &nbsp;
 * float ratio = parser.getFloatParameter("ratio", 1.0);
 * &nbsp;
 * int count = 0;
 * try {
 *   count = parser.getIntParameter("count");
 * }
 * catch (NumberFormatException e) {
 *   handleMalformedCount();
 * }
 * catch (ParameterNotFoundException e) {
 *   handleNoCount();
 * }
 * </pre></blockquote>
 *
 * There's also a capability to find out if any required parameters are
 * missing from a request:
 * <blockquote><pre>
 * ParameterParser parser = new ParameterParser(req);
 * String[] required = { "fname", "lname", "account" };
 * String[] missing = parser.getMissingParameters(required);
 * </pre></blockquote>
 *
 * @see com.oreilly.servlet.ParameterNotFoundException
 *
 * @author <b>Jason Hunter</b>, Copyright &#169; 1998, 1999
 * @version 1.1, 99/12/20, added getMissingParameters() method
 * @version 1.0, 98/09/18
 */
public class ParameterParser {

  private ServletRequest req;

  /**
   * Constructs a new ParameterParser to handle the parameters of the
   * given request.
   *
   * @param req the servlet request
   */
  public ParameterParser(ServletRequest req) {
    this.req = req;
  }

  /**
   * Gets the named parameter value as a String
   *
   * @param name the parameter name
   * @return the parameter value as a String
   * @exception ParameterNotFoundException if the parameter was not found
   * or was the empty string
   */
  public String getStringParameter(String name)
      throws ParameterNotFoundException {
    String[] values = req.getParameterValues(name);
    if (values == null)
      throw new ParameterNotFoundException(name + " not found");
    else if (values[0].length() == 0)
      throw new ParameterNotFoundException(name + " was empty");
    else
      return values[0];
  }

  /**
   * Gets the named parameter value as a String, with a default.
   * Returns the default value if the parameter is not found or 
   * is the empty string.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a String, or the default
   */
  public String getStringParameter(String name, String def) {
    try { return getStringParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a boolean
   *
   * @param name the parameter name
   * @return the parameter value as a boolean
   * @exception ParameterNotFoundException if the parameter was not found
   */
  public boolean getBooleanParameter(String name)
      throws ParameterNotFoundException {
    return new Boolean(getStringParameter(name)).booleanValue();
  }

  /**
   * Gets the named parameter value as a boolean, with a default.
   * Returns the default value if the parameter is not found.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a boolean, or the default
   */
  public boolean getBooleanParameter(String name, boolean def) {
    try { return getBooleanParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a byte
   *
   * @param name the parameter name
   * @return the parameter value as a byte
   * @exception ParameterNotFoundException if the parameter was not found
   * @exception NumberFormatException if the parameter value could not
   * be converted to a byte
   */
  public byte getByteParameter(String name)
      throws ParameterNotFoundException, NumberFormatException {
    return Byte.parseByte(getStringParameter(name));
  }

  /**
   * Gets the named parameter value as a byte, with a default.
   * Returns the default value if the parameter is not found or cannot
   * be converted to a byte.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a byte, or the default
   */
  public byte getByteParameter(String name, byte def) {
    try { return getByteParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a char
   *
   * @param name the parameter name
   * @return the parameter value as a char
   * @exception ParameterNotFoundException if the parameter was not found
   * or was the empty string
   */
  public char getCharParameter(String name)
      throws ParameterNotFoundException {
    String param = getStringParameter(name);
    if (param.length() == 0)
      throw new ParameterNotFoundException(name + " is empty string");
    else
      return (param.charAt(0));
  }

  /**
   * Gets the named parameter value as a char, with a default.
   * Returns the default value if the parameter is not found.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a char, or the default
   */
  public char getCharParameter(String name, char def) {
    try { return getCharParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a double
   *
   * @param name the parameter name
   * @return the parameter value as a double
   * @exception ParameterNotFoundException if the parameter was not found
   * @exception NumberFormatException if the parameter could not be converted
   * to a double
   */
  public double getDoubleParameter(String name)
      throws ParameterNotFoundException, NumberFormatException {
    return new Double(getStringParameter(name)).doubleValue();
  }

  /**
   * Gets the named parameter value as a double, with a default.
   * Returns the default value if the parameter is not found.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a double, or the default
   */
  public double getDoubleParameter(String name, double def) {
    try { return getDoubleParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a float
   *
   * @param name the parameter name
   * @return the parameter value as a float
   * @exception ParameterNotFoundException if the parameter was not found
   * @exception NumberFormatException if the parameter could not be converted
   * to a float
   */
  public float getFloatParameter(String name)
      throws ParameterNotFoundException, NumberFormatException {
    return new Float(getStringParameter(name)).floatValue();
  }

  /**
   * Gets the named parameter value as a float, with a default.
   * Returns the default value if the parameter is not found.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a float, or the default
   */
  public float getFloatParameter(String name, float def) {
    try { return getFloatParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a int
   *
   * @param name the parameter name
   * @return the parameter value as a int
   * @exception ParameterNotFoundException if the parameter was not found
   * @exception NumberFormatException if the parameter could not be converted
   * to a int
   */
  public int getIntParameter(String name)
      throws ParameterNotFoundException, NumberFormatException {
    return Integer.parseInt(getStringParameter(name));
  }

  /**
   * Gets the named parameter value as a int, with a default.
   * Returns the default value if the parameter is not found.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a int, or the default
   */
  public int getIntParameter(String name, int def) {
    try { return getIntParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a long
   *
   * @param name the parameter name
   * @return the parameter value as a long
   * @exception ParameterNotFoundException if the parameter was not found
   * @exception NumberFormatException if the parameter could not be converted
   * to a long
   */
  public long getLongParameter(String name)
      throws ParameterNotFoundException, NumberFormatException {
    return Long.parseLong(getStringParameter(name));
  }

  /**
   * Gets the named parameter value as a long, with a default.
   * Returns the default value if the parameter is not found.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a long, or the default
   */
  public long getLongParameter(String name, long def) {
    try { return getLongParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Gets the named parameter value as a short
   *
   * @param name the parameter name
   * @return the parameter value as a short
   * @exception ParameterNotFoundException if the parameter was not found
   * @exception NumberFormatException if the parameter could not be converted
   * to a short
   */
  public short getShortParameter(String name)
      throws ParameterNotFoundException, NumberFormatException {
    return Short.parseShort(getStringParameter(name));
  }

  /**
   * Gets the named parameter value as a short, with a default.
   * Returns the default value if the parameter is not found.
   * 
   * @param name the parameter name
   * @param def the default parameter value
   * @return the parameter value as a short, or the default
   */
  public short getShortParameter(String name, short def) {
    try { return getShortParameter(name); }
    catch (Exception e) { return def; }
  }

  /**
   * Determines which of the required parameters were missing from the
   * request.  Returns null if all the parameters are present.
   * 
   * @param an array of required parameters
   * @return an array of missing parameters, or null if none are missing
   */
  public String[] getMissingParameters(String[] required) {
    Vector missing = new Vector();
    for (int i = 0; i < required.length; i++) {
      String val = getStringParameter(required[i], null);
      if (val == null) {
        missing.addElement(required[i]);
      }
    }
    if (missing.size() == 0) {
      return null;
    }
    else {
      String[] ret = new String[missing.size()];
      missing.copyInto(ret);
      return ret;
    }
  }
}
